#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2021 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Created on Thu Jun 15 16:29:00 2021

WARNING: For simplicity and brevity of code this example does not show the
best way of handling the received data. Make sure to have a look at
example_buffered_data_callbacks3.py.

Test of the buffered-data callbacks interface.
This example performs the computation of a histogram from event data in
python code rather than letting the scTDC1 library do it.
(However, performing expensive computations in the 'on_data' callback
slows down the internal data processing of the scTDC1 library and is not
recommended).
Note that matplotlib functions should always be called from the main
thread, so they cannot be placed into the 'on_data' callback. Here, we defer
the visualization until measurements have been finished and the device has
been deinitialized.
"""

import scTDC
import time
import timeit
import numpy as np
from matplotlib import pyplot as plt
#

# -----------------------------------------------------------------------------
# example 2 of deriving from buffered_data_callbacks_pipe

NR_OF_MEASUREMENTS = 10

DATA_FIELD_SEL1 = \
      scTDC.SC_DATA_FIELD_DIF1 \
    | scTDC.SC_DATA_FIELD_DIF2 \
    | scTDC.SC_DATA_FIELD_TIME

class BufDataCB2(scTDC.buffered_data_callbacks_pipe):
    def __init__(self, lib, dev_desc,
                 data_field_selection=DATA_FIELD_SEL1,
                 max_buffered_data_len=500000,
                 dld_events=True):
        super().__init__(lib, dev_desc, data_field_selection,  # <-- mandatory!
                         max_buffered_data_len, dld_events)    # <-- mandatory!
        self.xy_histo = None
        self.xmax = None
        self.ymax = None
        self.total_events = 0
        self.buffer_meas = NR_OF_MEASUREMENTS - 1

    def on_data(self, d):
        # all numpy arrays in d will only be valid until we return from
        # this function! if you want to retain data for longer, you will
        # need numpy.ndarray.copy. For example calling 'd["dif1"].copy()'
        # returns a copied array which you can assign to a python variable
        # however, here we just retain processed results
        self.total_events += d["data_len"]
        print("data callback with {} events".format(d["data_len"]))
        print(" and {} start of measurements".format(len(d["som_indices"])))
        if d["data_len"] == 0:
          return
        # demonstrate that arbitrary transforms of the event data are possible
        # rotate the xy image by an angle around a center
        angle = 22.5
        binning = 4
        if self.xmax is None:
            self.xmax = np.max(d["dif1"])
            self.ymax = np.max(d["dif2"])
        imgdim = max(self.xmax, self.ymax)
        xcenter = self.xmax/2.0
        ycenter = self.ymax/2.0
        cosa = np.cos(angle / 180.0 * np.pi)
        sina = np.sin(angle / 180.0 * np.pi)
        xrot = cosa * (d["dif1"] - xcenter) - sina * (d["dif2"] - ycenter) + xcenter
        yrot = sina * (d["dif1"] - xcenter) + cosa * (d["dif2"] - ycenter) + ycenter
        # now build a 2D histogram of it (i.e. an image of the counts vs
        # detector position)
        # note: there is a python module named fast_histogram
        # (which you can install by calling python with the arguments
        #  "-m pip install --user fast_histogram")
        # fast_histogram is supposedly considerably faster than numpy's routine
        h,_xe,_ye = np.histogram2d(xrot, yrot, range=[[0,imgdim], [0,imgdim]],
                           bins=imgdim/binning)
        # (numpy.histogram2d returns a triple holding the result in element 0)
        # cache this histogram data and add histogram data of subsequent invocations
        if self.xy_histo is None:
            self.xy_histo = h
        else:
            self.xy_histo += h

    def on_end_of_meas(self):
        print("end of measurement")
        if self.buffer_meas > 0:
            self.buffer_meas -= 1 # count down
            return False          # don't trigger an on_data callback yet
        else:
            self.buffer_meas = NR_OF_MEASUREMENTS
            return True # last measurement, trigger on_data callback


# -----------------------------------------------------------------------------


def test2():
    device = scTDC.Device(autoinit=False)

    # initialize TDC --- and check for error!
    retcode, errmsg = device.initialize()
    if retcode < 0:
        print("error during init:", retcode, errmsg)
        return 0
    else:
        print("successfully initialized")

    # open a BUFFERED_DATA_CALLBACKS pipe
    bufdatacb = BufDataCB2(device.lib, device.dev_desc)
    start = timeit.default_timer()

    # do measurements
    for i in range(NR_OF_MEASUREMENTS):
        retcode = bufdatacb.start_measurement_sync(100)
        if retcode < 0:
            print(device.lib.sc_get_err_msg(retcode))
            break
    end = timeit.default_timer()
    print("\ntime elapsed : ", end-start, "s")

    time.sleep(0.1)
    # clean up
    bufdatacb.close() # closes the user callbacks pipe, method inherited from base class
    device.deinitialize()

    # our class still has the cached histogram data
    print("total events {}".format(bufdatacb.total_events))
    if bufdatacb.xy_histo is not None:
        plt.imshow(bufdatacb.xy_histo)
        plt.show()

if __name__ == "__main__":
    test2()
