#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Example showing how to set exposure parameters and how to start a measurement
with multiple camera frames and how to receive meta information and blob data
of the individual frames.
"""

import scTDC
import traceback

def example_camera_blobs():
    """Opens a camera frame pipe (for the meta data) and a blob pipe and starts
    one measurement, then reads the meta data + blobs of each frame and prints
    some of the blob positions. If meta data are required in addition to blob
    data (as done in this example), please pay attention to have each reading of
    the frames pipe followed by a reading of the blobs pipe such that they are
    kept synchronized in terms of the camera frame they return data for. For
    example, if you run into timeout while reading the frame pipe, it is best
    to repeat reading from the frame pipe until successful before following it
    by a single read of the blob pipe. If the pipes somehow should become
    unsynchronized, you can close and reopen both pipes, which discards all
    unread data."""
    FRAMES_COUNT = 5
    EXPOSURE_MICROSECS = 5000
    lib = scTDC.scTDClib()
    camera = scTDC.Camera(inifilepath="tdc_gpx3.ini", autoinit=False, lib=lib)
    v = lib.sc_tdc_config_get_library_version()

    print("scTDC library version {}.{}.{}".format(v[0], v[1], v[2]))
    ret, msg = camera.initialize()
    print("initialization result: {} {}".format(ret, msg))

    if ret >= 0:
        try:
            camera.set_fanspeed(150)
            camera.set_exposure_and_frames(EXPOSURE_MICROSECS, FRAMES_COUNT)
            camera.set_region_of_interest(0, 1940, 0, 1440)
            camera.set_smoother_bit_shifts(2, 2) # overrides settings from ini file
            camera.set_smoother_masks_square(2, 2) # overrides settings from ini file
            #camera.set_blob_mode(1, 3) # overrides settings from ini file
            print("get ROI", camera.get_region_of_interest())
            pipeid_meta, pipe_meta = camera.add_frame_pipe()
            pipeid_blobs, pipe_blobs = camera.add_blobs_pipe()
            camera.do_measurement(1) # the argument value here is arbitrary
            for i in range(FRAMES_COUNT):
                meta_data = pipe_meta.read(timeout_ms = 1000)[0]
                if isinstance(meta_data, int):
                    print("error reading meta data")
                    continue
                print("meta data", meta_data)
                blobs = pipe_blobs.read(timeout_ms = 1000)
                if isinstance(blobs[0], int):
                    if blobs[0]==0:
                        print("no blobs")
                    else:
                        print("error reading blobs", blobs[1])
                else:
                    print_blob_count = min(5, len(blobs)) # print at most 5 blob positions
                    print("number of blobs", len(blobs))
                    print("coordinates of first {0} blobs".format(print_blob_count))
                    for blobidx in range(print_blob_count):
                        print(" ", blobs[blobidx])
            pipe_meta.close()
            pipe_blobs.close()
        finally:
            camera.deinitialize()

if __name__ == "__main__":
    example_camera_blobs()