# -*- coding: utf-8 -*-
"""
Copyright 2019 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Test end-of-measurement callback.

"""



def test_end_of_measurement_callback():
    """ Performs 1 measurement asynchronously and waits until the callback
    for end of measurement arrives. (prints statistics data, afterwards). """
    import scTDC
    import threading
    
    device = scTDC.Device(autoinit=False)
    retcode, errmsg = device.initialize()
    if retcode < 0:
        print("Error during initialization : ({}) {}".format(errmsg, retcode))
        return
    
    pipeid, pipe = device.add_statistics_pipe()

    # prepare condition variable for waiting on end of measurement
    cv = threading.Condition()
    measurement_finished = [False] # use a list because it is mutable
    # define and register an end-of-measurement callback
    # with real hardware, you often get two callbacks at the end of measurement
    # once with code 4, then with code 1. Here, we filter out the code 4
    # (CBR_EARLY_NOTIF)
    def _on_eom(reason): # construct a closure
        try:
            print(scTDC.CBR_DICT[reason])
        except KeyError:
            print("Got end-of-measurement callback with code", reason)
        if reason!=scTDC.CBR_EARLY_NOTIF:
            with cv:
                measurement_finished[0] = True
                cv.notify()
    device.add_end_of_measurement_callback(_on_eom)

    # start the measurement
    with cv: # locks the lock of the condition variable
        print("Starting a measurement")
        retcode, errmsg = device.do_measurement(time_ms=100, synchronous=False)
        #device.interrupt_measurement() # can do this to test CBR_USER_ABORT
        if retcode < 0:
            print("Error while starting measurement : ({}) {}".format(
                errmsg, retcode))
        else:
            # the while loop protects against spurious wake-ups
            while not measurement_finished[0]:
                cv.wait()

    # --- the following is not specific to the end-of-meas. callback technique
    # collect statistics data and print it
    stat_data = pipe.get_buffer_copy()
    print("non-zero raw counts by channel:")
    for i in range(16):
        r = stat_data.counts_read[i]
        if r > 0:
            print(" {:2} {:15}".format(i, r))
    maxcounts = max([stat_data.counts_read[i] for i in range(16)])
    print("highest of the raw counts :", maxcounts)
    print("events found :", stat_data.events_found[0])
    if maxcounts > 0:
        print("ratio of found events to highest raw count : {:.3g}".format( 
              stat_data.events_found[0]/maxcounts))

    
    device.deinitialize()

if __name__ == "__main__":
    test_end_of_measurement_callback()
