#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright 2023 Surface Concept GmbH

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

-------------------------------------------------------------------------------

Example for a PyQt5-based GUI which uses a histogram pipe for stand-alone TDCs
(however, the pipe type can be easily changed for DLD applications, as well).
The example demonstrates how to start a measurement asynchronously (in the
background, without blocking the GUI thread) and how to correctly implement
actions that need to happen at the end of the measurement, using an
"end of measurement" callback from the scTDC library. Since this callback
executes in a non-GUI thread, the first step is to notify the GUI thread such
that all data processing, scheduling of further measurements, etc. can be run in
the GUI thread.

The application offers controls to initialize/deinitialize the device, start
a single measurement, run a sequence of measurements.

Tested with PyQt5 version 5.12.3.
"""

import time
import traceback
from PyQt5.QtWidgets import QApplication, QWidget, QPushButton, QSpinBox, \
  QVBoxLayout, QHBoxLayout, QLabel, QCheckBox, QStatusBar, QMainWindow
import PyQt5.QtCore as QtCore
import numpy as np
import scTDC

MINIMUM_SIZE_X = 400
MINIMUM_SIZE_Y = 150

class MainWidget(QWidget):
  def __init__(self, parent=None, show_message=None):
    """constructs a MainWidget object

    :param parent: a QWidget parent, defaults to None
    :type parent: QWidget, optional
    :param status_message: a function that takes strings containing status
      messages, defaults to None
    :type status_message: str, optional
    """
    super().__init__(parent)
    if show_message is not None:
      self.show_message = show_message
    else:
      def show_message_ignore(*args):
        pass
      self.show_message = show_message_ignore
    self.bu_init = QPushButton('Init')
    self.bu_deinit = QPushButton('Deinit')
    self.l_init_status = QLabel('not initialized')
    self.l_init_status.setStyleSheet("background-color: #FFCCCC; padding: 5px;")
    self.l_init_status.setMinimumWidth(self.fontMetrics().averageCharWidth()*12)
    self.bu_start_measurement = QPushButton('Start')
    self.bu_start_measurement.setEnabled(False)
    self.l_measurement_status = QLabel('idle')
    self.l_measurement_status.setMinimumWidth(self.fontMetrics().averageCharWidth()*12)
    self.l_measurement_status.setStyleSheet('background-color: #CCCCCC; padding: 5px;')
    self.cb_sequence = QCheckBox('Sequence')
    self.cb_sequence.setEnabled(False)
    self.l_sequence_count = QLabel("0")
    self.l_sequence_count.setMinimumWidth(8*self.fontMetrics().averageCharWidth())
    self.sequence_count = 0
    self.sb_exposure = QSpinBox()
    self.sb_exposure.setRange(1, 2147483647)
    self.sb_exposure.setValue(200)
    self.l_histo_total_counts = QLabel('0')
    self.l_histo_total_counts.setMinimumWidth(self.fontMetrics().averageCharWidth()*16)
    self.l_histo_total_counts.setStyleSheet('background-color: #CCCCCC; padding: 5px;')
    ### LAYOUT ###
    layout = QVBoxLayout()
    row1 = QHBoxLayout()
    row1.addWidget(self.bu_init)
    row1.addStretch(1)
    row1.addWidget(self.bu_deinit)
    row1.addStretch(1)
    row1.addWidget(self.l_init_status)
    row1.addStretch(5)
    row2 = QHBoxLayout()
    row2.addWidget(QLabel("measurement:"))
    row2.addWidget(self.bu_start_measurement)
    row2.addStretch()
    row2.addWidget(self.l_measurement_status)
    row2.addStretch()
    row2.addWidget(self.cb_sequence)
    row2.addWidget(self.l_sequence_count)
    row3 = QHBoxLayout()
    row3.addWidget(QLabel("exposure time (ms)"))
    row3.addWidget(self.sb_exposure)
    row3.addStretch() # for left alignment
    row4 = QHBoxLayout()
    row4.addWidget(QLabel("total counts in histogram:"))
    row4.addWidget(self.l_histo_total_counts)
    layout.addLayout(row1)
    layout.addLayout(row2)
    layout.addLayout(row3)
    layout.addLayout(row4)
    layout.addStretch()
    self.setLayout(layout)
    self.show()
    ### SIGNALS ###
    self.bu_init.clicked.connect(self.slot_bu_init)
    self.bu_deinit.clicked.connect(self.slot_bu_deinit)
    self.bu_start_measurement.clicked.connect(self.slot_bu_start_measurement)
    self.cb_sequence.stateChanged.connect(self.slot_cb_sequence)
    ### HARDWARE-related
    self.exposure_time_ms = 200
    self.tdc_histo_pipe = None
    self.tdc_histo_pipe_id = None
    self.device = None
    self.setup_hardware()

  def setup_hardware(self):
    try:
      self.device = scTDC.Device(autoinit=False)
      ### Register Callback
      # The registered callback is not affected by initialization or
      # deinitialization, so registration is only required once at application
      # startup.
      callback_id = self.device.add_end_of_measurement_callback(
        self.on_end_of_measurement)
      if callback_id < 0:
        self.show_message(
          "Callback registration failed: ({})".format(callback_id))
    except OSError as e:
      print(e)
      self.show_message("Unable to load library scTDC")
      self.setEnabled(False)

  def slot_bu_init(self, *args):
    if self.device.is_initialized():
      return
    ### 1 Initialize
    retcode, errmsg = self.device.initialize()
    if retcode < 0:
      self.show_message("Init failed: ({}) {}".format(errmsg, retcode))
      return
    ### 2 Create Pipe
    pipeid, pipe = self.device.add_tdc_histo_pipe(depth=scTDC.BS16, channel=-1,
      modulo=0, binning=1<<14, offset=0, size=128)
    if pipeid < 0:
      self.show_message("Pipe creation failed: {}".format(pipeid))
      self.tdc_histo_pipe = None
      self.tdc_histo_pipe_id = None
       # initialization did not fully succeed, undo previous steps:
      self.device.deinitialize()
      return
    self.tdc_histo_pipe = pipe
    self.tdc_histo_pipe_id = pipeid
    ### 3 GUI cosmetics
    self.l_init_status.setText("initialized")
    self.l_init_status.setStyleSheet("background-color: #CCFFCC; padding: 5px;")
    self.bu_start_measurement.setEnabled(True)
    self.cb_sequence.setEnabled(True)
    self.bu_init.setEnabled(False)
    self.bu_deinit.setEnabled(True)

  def slot_bu_deinit(self, *args):
    if self.tdc_histo_pipe_id is not None:
      self.tdc_histo_pipe.close()
    if self.device is not None and self.device.is_initialized():
      self.device.deinitialize()
    self.l_init_status.setText("not initialized")
    self.l_init_status.setStyleSheet("background-color: #FFCCCC; padding: 5px;")
    self.bu_start_measurement.setEnabled(False)
    self.cb_sequence.setEnabled(False)
    self.bu_init.setEnabled(True)
    self.bu_deinit.setEnabled(False)


  def slot_bu_start_measurement(self, *args):
    self.tdc_histo_pipe.clear()
    self.l_measurement_status.setText("measuring...")
    retcode, errmsg = self.device.do_measurement(
      time_ms=self.sb_exposure.value(), synchronous=False)
    if retcode < 0:
      self.show_message(errmsg, 5000)
      self.l_measurement_status.setText("idle")

  def slot_cb_sequence(self, *args):
    # When user checks the 'Sequence' checkbox, kick off the sequence by
    # starting its first measurement. Follow-up measurements are triggered from
    # the handle_end_of_measurement function, which checks whether the 'Sequence'
    # checkbox is checked. The user can stop the sequence by unchecking the
    # checkbox.
    if self.cb_sequence.isChecked():
      self.sequence_count = 0
      self.l_sequence_count.setText("{}".format(self.sequence_count))
      self.slot_bu_start_measurement()

  def on_end_of_measurement(self, reason):
    """This function is registered during initialization to be called by the
    scTDC library when the end of measurement is reached and executes in a
    non-GUI thread.

    :param reason: one of 1,2,3,4. 4 happens additionally to other reasons in a
      separate invocation of this function to notify that the hardware is idle
      but data transfer to the PC is still in process. 1 is regular completion.
      2 and 3 are for interrupted measurements by user or by other conditions.
    :type reason: int
    """
    # In order to transfer the information about the end of measurement to the
    # GUI thread, we use a mechanism by Qt to call a slot indirectly, a
    # thread-safe way to tell the GUI thread that it should execute the
    # 'handle_end_of_measurement' function as soon as possible. The
    # QMetaObject.invokeMethod also decouples the chain of function calls, so
    # this function can return before the 'handle_end_of_measurement' is
    # actually executed. This is important for the scTDC library to transition
    # back to the idle state where new measurements can be started.
    # Furthermore, it is unsafe to interact with the GUI from a non-GUI thread,
    # directly.
    QtCore.QMetaObject.invokeMethod(
      self,
      'handle_end_of_measurement',
      QtCore.Qt.QueuedConnection,
      QtCore.Q_ARG(int, reason))

  @QtCore.pyqtSlot(int)
  def handle_end_of_measurement(self, reason):
    """This function is designed to run in the GUI thread and performs any
    actions required at the end of a measurement, such as data readout, data
    processing, start of follow-up measurements, etc."""
    if reason == scTDC.CBR_COMPLETE: # CBR_COMPLETE: regular completion
      self.process_data()
    if reason == scTDC.CBR_COMPLETE and self.cb_sequence.isChecked():
      self.sequence_count += 1
      self.l_sequence_count.setText("{}".format(self.sequence_count))
      self.tdc_histo_pipe.clear()
      self.start_next_measurement()
    elif reason != scTDC.CBR_EARLY_NOTIF:
      self.l_measurement_status.setText("idle")

  def start_next_measurement(self):
    for i in range(5): # retry 5 times
      retcode, errmsg = self.device.do_measurement(
        time_ms=self.sb_exposure.value(), synchronous=False)
      if retcode != -11:
        if retcode != 0: # if error other than "NOT READY": show message
          self.show_message(errmsg)
        break # if success or error other than "NOT READY": exit loop
      time.sleep(0.01) # wait before trying again

  def process_data(self):
    try:
      data = self.tdc_histo_pipe.get_buffer_view()
      self.l_histo_total_counts.setText("{}".format(np.sum(data)))
    except:
      traceback.print_exc()

if __name__ == "__main__":
  app = QApplication([])
  mainwindow = QMainWindow()
  mainwindow.setWindowTitle("Example TDC (PyQt5)")
  mainwindow.setMinimumSize(MINIMUM_SIZE_X, MINIMUM_SIZE_Y)
  mainwindow.status_bar = QStatusBar()
  mainwindow.setStatusBar(mainwindow.status_bar)
  mainwindow.statusBar().setStyleSheet("border-top: 2px inset #CCCCCC;")

  w = MainWidget(mainwindow, mainwindow.status_bar.showMessage)

  mainwindow.setCentralWidget(w)
  mainwindow.show()
  app.exec()
